/************************************************************************
 *   IRC - Internet Relay Chat, admin/m_add.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ccr.h"
#include "channel.h"
#include "class.h"
#include "client.h"
#include "irc_string.h"
#include "ircd.h"
#include "numeric.h"
#include "s_conf.h"
#include "s_log.h"
#include "send.h"
#include "struct.h"
#include "umodes.h"
#include "memory.h"
#include "msg.h"
#include "handlers.h"
#include "database/database.h"
#include "varparse.h"
#include "flagparse.h"

#include <arpa/inet.h>

int m_add(struct Client *, struct Client *, int, char **);

struct Message add_msgtab = {
  "ADD", 0, 0, 0, MFLG_SLOW, 0,
  {m_unregistered, m_add, m_add},
  NULL, NULL
};

var_handler_func m_add_ccr_allow, m_add_ccr_deny, m_add_drop;

static struct variable_tier add_ccr_parse_tree[] = {
  {"ALLOW", 0, NULL, m_add_ccr_allow},
  {"DENY",  0, NULL, m_add_ccr_deny},
  {NULL, 0, NULL, NULL}
};

static struct variable_tier add_parse_tree[] = {
  {"CCR", 0, add_ccr_parse_tree, NULL},
  {"DROP", 0, NULL, m_add_drop},
  {NULL, 0, NULL, NULL}
};

static struct flag_def ccr_flags[] = {
  {"no-tilde", CCONNECT_FLAGS_NO_TILDE},
  {"need-ident", CCONNECT_FLAGS_NEED_IDENT},
  {"allow-bots", CCONNECT_FLAGS_ALLOW_BOTS},
  {NULL, NULL_FLAG}
};

int
m_add(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
  return variable_parse(cptr, sptr, parc, parv, add_parse_tree, "ADD");
}

/* NOTE: deny ccrs take effect immediately. allow ccrs take effect on
 *  the next connection. this may change if absolutely necessary, but for
 *  now I don't want to write that horribly complex routine.
 *  -- asuffield
 */

int
m_add_ccr_allow(struct Client *cptr, struct Client *sptr, int parc, char *parv[], char *varparv[])
{
  struct client_connect_record ccr;
  int i;
  struct Client *acptr, *next_cptr;

  if (!HasUmode(sptr, UMODE_KILL))
    {
      if (SeesOperMessages(sptr))
	sendto_one(sptr, get_str(STR_NEEDPRIV), me.name, parv[0],
		   user_mode_char[UMODE_KILL]);
      else
	sendto_one(sptr, form_str(ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
    }

  if (parc < 3)
    {
      if (!IsServer(sptr))
	sendto_one(sptr, form_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "ADD CCR.ALLOW");
      return 0;
    }

  memset(&ccr, 0, sizeof(ccr));
  ccr.host = parv[2];
  ccr.when = CurrentTime;
  ccr.placed_by = sptr->name;

  for (i = 3; i < parc; i++)
    {
      if (!irccmp(parv[i], "no-tilde"))
	ccr.flags &= CCONNECT_FLAGS_NO_TILDE;
      else if (!irccmp(parv[i], "need-ident"))
	ccr.flags &= CCONNECT_FLAGS_NEED_IDENT;
      else if (!irccmp(parv[i], "allow-bots"))
	ccr.flags &= CCONNECT_FLAGS_ALLOW_BOTS;
      else if (i < (parc - 1))
	{
	  if (!irccmp(parv[i], "until"))
	    {
	      ccr.until = atoi(parv[++i]);
	    }
	  else if (!irccmp(parv[i], "for"))
	    {
	      ccr.until = CurrentTime + timestrtol(parv[++i]);
	    }
	  else if (!irccmp(parv[i], "priority"))
	    {
	      ccr.priority = atoi(parv[++i]);
	    }
	  else if (!irccmp(parv[i], "by"))
	    {
	      if (IsServer(cptr))
		ccr.placed_by = parv[++i];
	      else
		i++;
	    }
	  else if (!irccmp(parv[i], "class"))
	    {
	      ccr.class = atoi(parv[++i]);
	    }
	  else if (!irccmp(parv[i], "password"))
	    {
	      ccr.passwd = parv[++i];
	    }
	  else if (!irccmp(parv[i], "spoof"))
	    {
	      ccr.spoof_host = parv[++i];
	    }
	  else if (!irccmp(parv[i], "because"))
	    {
	      ccr.reason = parv[++i];
	    }
	}
    }

  /* OK, that should do for the parsing. Add the beggar... */
  store_allow_ccr(&ccr);

  /* I have got to think of a better way to do this */
  sendto_serv_butone(cptr, ":%s ADD CCR.ALLOW %s until %.1lu priority %.1u class %.1u by %s"
		     " %s %s %s %s %s %s %s %s :%s",
		     sptr->name, ccr.host, ccr.until, ccr.priority,
		     ccr.class, ccr.placed_by,
		     ccr.spoof_host ? "spoof" : "", ccr.spoof_host ? ccr.spoof_host : "",
		     ccr.passwd ? "passwd" : "", ccr.passwd ? ccr.passwd : "",
		     ccr.flags & CCONNECT_FLAGS_NO_TILDE ? "no-tilde" : "",
		     ccr.flags & CCONNECT_FLAGS_NEED_IDENT ? "need-ident" : "",
		     ccr.flags & CCONNECT_FLAGS_ALLOW_BOTS ? "allow-bots" : "",
		     ccr.reason ? "because" : "", ccr.reason ? ccr.reason : "");

  if (MyClient(sptr))
    sendto_one(sptr, form_str(RPL_CCONNECT), me.name, sptr->name,
	       ccr.host, ccr.priority, ccr.placed_by, ccr.when,
	       ccr.until, flag_string(ccr.flags, ccr_flags),
	       ccr.deleted_by ? ccr.deleted_by : "-",
	       ccr.time_deleted,
	       ccr.reason);

  return 0;
}

int
m_add_ccr_deny(struct Client *cptr, struct Client *sptr, int parc, char *parv[], char *varparv[])
{
  struct client_connect_record ccr;
  int i;
  struct Client *acptr, *next_cptr = NULL;

  if (!HasUmode(sptr, UMODE_KILL))
    {
      if (SeesOperMessages(sptr))
	sendto_one(sptr, get_str(STR_NEEDPRIV), me.name, parv[0],
		   user_mode_char[UMODE_KILL]);
      else
	sendto_one(sptr, form_str(ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
    }

  if (parc < 3)
    {
      if (!IsServer(sptr))
	sendto_one(sptr, form_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "ADD CCR.DENY");
      return 0;
    }

  memset(&ccr, 0, sizeof(ccr));
  ccr.host = parv[2];
  ccr.when = CurrentTime;
  ccr.placed_by = sptr->name;

  for (i = 3; i < (parc - 1); i++)
    {
      if (!irccmp(parv[i], "until"))
	{
	  ccr.until = atoi(parv[++i]);
	}
      else if (!irccmp(parv[i], "for"))
	{
	  ccr.until = CurrentTime + timestrtol(parv[++i]);
	}
      else if (!irccmp(parv[i], "priority"))
	{
	  ccr.priority = atoi(parv[++i]);
	}
      else if (!irccmp(parv[i], "because"))
	{
	  ccr.reason = parv[++i];
	}
      else if (!irccmp(parv[i], "by"))
	{
	  if (IsServer(cptr))
	    ccr.placed_by = parv[++i];
	  else
	    i++;
	}
    }

  /* Safety checks */
  {
    char ch, *p = ccr.host;
    int nonwild = 0;

    while ((ch = *p++))
      if (!IsKWildChar(ch))
	if (++nonwild >= NONWILDCHARS)
	  /* Stop as soon as we have enough */
	  break;

    if (nonwild < NONWILDCHARS)
      {
	if (IsClient(sptr))
	  sendto_one(sptr, get_str(STR_NEEDMORENONWILDS), me.name, sptr->name,
		     NONWILDCHARS);
	return -1;
      }
  }

  /* OK, that should do for the parsing. Add the beggar... */
  store_deny_ccr(&ccr);

  /* ...and kill some local lusers */
  for (acptr = local_cptr_list; (acptr = next_cptr);)
    {
      if (acptr) /* Sigh */
	{
	  /* Grab the next one now, we might be going to delete this one */
	  next_cptr = acptr->next_local_client;
	  if (IsPerson(acptr) && (ccr.priority >= acptr->ccr_priority))
	    {
	      sendto_ops_flag(UMODE_SKILL, "%s confirming kill of %s by %s (%s)",
			      me.name, acptr->name, sptr->name,
			      ccr.reason ? ccr.reason : "no reason");
	      sendto_one(cptr, form_str(ERR_YOUREBANNEDCREEP),
			 me.name, cptr->name, "K-lined");
	      exit_client(acptr, acptr, sptr, "K-lined");
	    }
	}
    }

  sendto_serv_butone(cptr, ":%s ADD CCR.ALLOW %s until %.1lu priority %.1u %s %s by %s",
		     sptr->name, ccr.host, ccr.until, ccr.priority,
		     ccr.reason ? "because" : "", ccr.reason ? ccr.reason : "",
		     ccr.placed_by);

  if (MyClient(sptr))
    sendto_one(sptr, form_str(RPL_CCONNECT), me.name, sptr->name,
	       ccr.host, ccr.priority, ccr.placed_by, ccr.when,
	       ccr.until, ",",
	       ccr.deleted_by ? ccr.deleted_by : "-",
	       ccr.time_deleted,
	       ccr.reason);
  return 0;
}

int
m_add_drop(struct Client *cptr, struct Client *sptr, int parc, char *parv[], char *varparv[])
{
  struct ip_drop_record idr;
  const char *host;
  int i, cidr;

  memset(&idr, 0, sizeof(idr));

  if (!HasUmode(sptr, UMODE_KILL))
    {
      if (SeesOperMessages(sptr))
	sendto_one(sptr, get_str(STR_NEEDPRIV), me.name, parv[0],
		   user_mode_char[UMODE_KILL]);
      else
	sendto_one(sptr, form_str(ERR_NOPRIVILEGES), me.name, parv[0]);
      return 0;
    }

  if (parc < 3)
    {
      if (!IsServer(sptr))
	sendto_one(sptr, form_str(ERR_NEEDMOREPARAMS),
		   me.name, parv[0], "ADD DROP");
      return 0;
    }

  host = parv[2];

  if (!is_address(host, &idr.ip, &idr.ip_mask))
    {
      struct Client *acptr;

      if (!(acptr = find_chasing(sptr, parv[1], NULL)))
	return 0;

      if (!acptr->user)
	return 0;

      if (IsServer(acptr))
	{
	  sendto_one(sptr,
		     ":%s NOTICE %s :Can't DLINE a server, silly",
		     me.name, parv[0]);
	  return 0;
	}

      if (IsElined(acptr))
	{
	  sendto_one(sptr,
		     ":%s NOTICE %s :%s is E-lined", me.name, parv[0],
		     acptr->name);
	  return 0;
	}

      host = inetntoa((const char*)&acptr->ip.s_addr);

      idr.ip_mask = 0xFFFFFFFFL;
      idr.ip = ntohl(acptr->ip.s_addr);
    }

  idr.placed_by = sptr->name;
  idr.placed_at  = CurrentTime;

  for (i = 3; i < (parc - 1); i++)
    {
      if (!irccmp(parv[i], "exception"))
	idr.exception = 1;
      else if (!irccmp(parv[i], "until"))
	{
	  idr.until = atoi(parv[++i]);
	}
      else if (!irccmp(parv[i], "for"))
	{
	  idr.until = CurrentTime + timestrtol(parv[++i]);
	}
      else if (!irccmp(parv[i], "because"))
	{
	  idr.reason = parv[++i];
	}
      else if (!irccmp(parv[i], "mask"))
	{
	  struct in_addr addr;
	  if (inet_aton(parv[++i], &addr))
	    idr.ip_mask = ntohl(addr.s_addr);
	}
      else if (!irccmp(parv[i], "by"))
	{
	  if (IsServer(cptr))
	    idr.placed_by = parv[++i];
	  else
	    i++;
	}
    }

  add_ip_drop(&idr);

  {
    /* IP address in network form */
    unsigned long int nip = htonl(idr.ip), nmask = htonl(idr.ip_mask);
    char ip_buf[33], mask_buf[33];
    strncpy_irc(ip_buf, inetntoa((char *)&nip), 32);
    strncpy_irc(mask_buf, inetntoa((char *)&nmask), 32);
    sendto_serv_butone(cptr, ":%s ADD DROP %s mask %s until %.1lu by %s %s %s :%s",
		       sptr->name, ip_buf, mask_buf, idr.until,
		       idr.placed_by, idr.exception ? "exception" : "",
		       idr.reason ? "because" : "", idr.reason ? idr.reason : "");
  }

  return 0;
}
